/*
	This is a ACLfs client.

	It maps various filesystems into the client home/aclfs directory.
	It deals with the aclfsd server to gain access to files. The later
	implement (may implement) sophisticated access control using
	ACL (Access Control List) instead of the limited Unix 9 bits
	permission scheme.
*/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <wait.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <dirent.h>
#include <limits.h>
#include <unistd.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <utime.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/socket.h>
#include <sys/un.h>
#include "aclfs.h"
#include "remaudio.h"
#include "dispatch.h"

#ifndef SCM_CREDENTIALS
	#define SCM_CREDENTIALS 2
	// Struct not available in headers yet, but supported by the kernel
	struct ACL_UCRED {
		int32_t	pid;
		int32_t	uid;
		int32_t	gid;
	};
#else
	typedef struct ucred ACL_UCRED;
#endif

extern "C" char *stpcpy (char *, const char *);

struct ACLFS_CONF {
	char *volume;
	char *realvolume;
};

static int nbdef=0;
static ACLFS_CONF *conf;

static void aclfs_readconf ()
{
	if (conf == NULL){
		char path[PATH_MAX];
		snprintf (path,sizeof(path)-1,"%s/.aclfs",_dispatch_gethome());
		FILE *fin = fopen (path,"r");
		if (fin != NULL){
			char buf[1000];
			conf = (ACLFS_CONF*)malloc (100*sizeof(ACLFS_CONF));
			while (fgets(buf,sizeof(buf)-1,fin)!=NULL){
				if (buf[0] != '\0'){
					char *pt = buf;
					while (*pt > ' ') pt++;
					if (isspace(*pt)){
						*pt ++ = '\0';
						while (isspace(*pt)) pt++;
						char *end = pt + strlen(pt)-1;
						if (*end == '\n') *end = '\0';
						conf[nbdef].volume = strdup(buf);
						conf[nbdef].realvolume = strdup(pt);
						nbdef++;
					}
				}
			}
			fclose (fin);
		}
	}
}

struct ACLFS_INFO{
	char volume[PATH_MAX];		// Sub-directory in $HOME/n
	char relpath[PATH_MAX];		// relative path in $HOME/n/volume/
	char virtpath[PATH_MAX];	// absolute virtual path
	char realpath[PATH_MAX];	// Translated path
};

/*
	Check if a volume is configured for the personal auto mounter
*/
static int aclfs_chkconf (ACLFS_INFO &info)
{
	aclfs_readconf();
	int ret = -1;
	for (int i=0; i<nbdef; i++){
		if (strcmp(info.volume,conf[i].volume)==0){
			sprintf (info.realpath,"/%s%s",conf[i].realvolume
				,info.relpath);
			ret = 0;
			break;
		}
	}
	return ret;
}

static int acl_fd=-1;	// Socket to do request

/*
	Wait to receive a file handle
	Return the file handle or -1.
*/
static int aclfs_receivefd()
{
	int ret = -1;
	char bufctl[sizeof(struct cmsghdr)+1*sizeof(int)];
	memset (bufctl,-1,sizeof(bufctl));
	struct iovec iov[1];
	char buf[1000];
	iov[0].iov_base = buf;
	iov[0].iov_len = sizeof(buf);
	struct msghdr h;
	h.msg_name = 0;
	h.msg_namelen = 0;
	h.msg_iov = iov;
	h.msg_iovlen = 1;
	h.msg_control = bufctl;
	h.msg_controllen = sizeof(bufctl);
	struct cmsghdr *c = (struct cmsghdr*) bufctl;
	c->cmsg_len = sizeof(bufctl);
	//c->cmsg_level = SOL_SOCKET;
	//c->cmsg_type = SCM_RIGHTS;
	h.msg_flags = 0;
	int len = recvmsg (acl_fd,&h,0);
	if (len<=0){
		perror ("recvmsg");
	}else{
		#if 0
			fprintf (stderr,"message recu flags %d len %d %d %u\n"
				,h.msg_flags,len
				,h.msg_controllen,sizeof(bufctl));
			fprintf (stderr,"iov.len %d\n",iov[0].iov_len);
			buf[len] = '\0';
			fprintf (stderr,"buf = :%s:\n",buf);
			fprintf (stderr,"msg_len %d msg_level %d msg_type %d\n"
				,c->cmsg_len
				,c->cmsg_level
				,c->cmsg_type);
		#endif
		/*
			On kernel 2.2, we can check c->cmsg_type and c->cmsg_level
			to find out if there is a file handle

			On kernel 2.0, we can only tell if there is
			a descriptor if it is != -1
			if (c->cmsg_level == SOL_SOCKET
				&& c->cmsg_type == SCM_RIGHTS){
		*/
		int *fd = (int*)CMSG_DATA(c);
		//fprintf (stderr,"fd passing %d\n",fd[0]);
		ret = fd[0];
	}
	return ret;
}

/*
	Wait for a confirmation status.
	Return -1 if the original request failed.
*/
static int aclfs_receive()
{
	int ret = -1;
	union {
		char buf[100];
		ACL_REPLY rep;
	}u;
	int len = read (acl_fd,&u,sizeof(u));
	if (len == sizeof(ACL_REPLY)){
		if (u.rep.status == ACL_STATUS_OK){
			ret = 0;
		}else{
			errno = u.rep.errcode;
		}
	}
	return ret;
}

/*
	Wait for a confirmation of a stat request.
	Return -1 if the original request failed.
*/
static int aclfs_receive_stat (struct stat &buf)
{
	int ret = -1;
	struct {
		ACL_REPLY rep;
		struct stat st;
	}u;
	int len = read (acl_fd,&u,sizeof(u));
	memset (&buf,0,sizeof(struct stat));
	if (len == sizeof(u)){
		errno = u.rep.errcode;
		if (u.rep.status == ACL_STATUS_STAT){
			memcpy (&buf,&u.st,sizeof(struct stat));
			ret = 0;
		}
	}
	return ret;
}

static int aclfs_connect ()
{
	int ret = -1;
	if (acl_fd != -1){
		ret = 0;
	}else{
		int fd =  socket (AF_UNIX,SOCK_STREAM,0);
		if (fd == -1){
			perror("socket client");
		}else{
			struct sockaddr_un un;
			un.sun_family = AF_UNIX;
			strcpy (un.sun_path,ACLFS_SOCKET);
			int s = connect(fd,(struct sockaddr*)&un,sizeof(un));
			if (s == -1){
				perror("connect");
			}else{
				acl_fd = fd;
				ret = 0;
			}
		}
	}
	return ret;
}

/*
	Translate a logical path if it is part of the personnal mount area
	($HOME/n)
*/
static int aclfs_translate (
	const char *path,
	ACLFS_INFO &info,
	ACL_REQUEST &req,
	int &len_request)
{
	const char *home = _dispatch_gethome();
	char pathn[PATH_MAX];
	int lenn = snprintf (pathn,sizeof(pathn)-1,"%s/acl/",home);
	strcpy (info.realpath,path);
	if (path[0] != '/'){
		const char *curpath = _dispatch_getcwd();
		snprintf (info.virtpath,sizeof(info.virtpath)-1,"%s/%s",curpath,path);
		path = info.virtpath;
	}else{
		strcpy (info.virtpath,path);
	}
	int ret = -1;

	//printf ("pathn = :%s: <> :%s:\n",pathn,path);
	if (strncmp(pathn,path,lenn)==0){
		path += lenn;
		char *volume = info.volume;
		while (*path != '\0' && *path != '/'){
			*volume++ = *path++;
		}
		*volume = '\0';
		strcpy (info.relpath,path);
		//printf ("volume :%s: :%s:\n",info.volume,info.relpath);
		if (aclfs_chkconf (info)!=-1){
			if (aclfs_connect() != -1){
				char *end = stpcpy (req.paths,info.realpath) + 1;
				len_request = (int)(end-(char*)&req);
				req.mode = 0;
				req.perm = 0;
				ret = 0;
			}
		}
	}
	return ret;
}

static void aclfs_sendreq (ACL_REQUEST &req, int len)
{
	char bufctl[sizeof(struct cmsghdr)+sizeof(ACL_UCRED)];
	struct cmsghdr *c = (struct cmsghdr*)bufctl;
	c->cmsg_len = CMSG_LEN(sizeof(ACL_UCRED));
	c->cmsg_level = SOL_SOCKET;
	c->cmsg_type = SCM_CREDENTIALS;
	ACL_UCRED *uc = (ACL_UCRED*)CMSG_DATA(c);
	uc->pid = getpid();
	uc->uid = getuid();
	uc->gid = getgid();


	struct msghdr h;

	struct iovec iov[1];
	iov[0].iov_base = &req;
	iov[0].iov_len = len;

	h.msg_name = 0;
	h.msg_namelen = 0;
	h.msg_iov = iov;
	h.msg_iovlen = 1;
	h.msg_control = (void*)bufctl;
	h.msg_controllen = sizeof(bufctl);
	h.msg_flags = 0;
	int ret = sendmsg (acl_fd,&h,0);
	if (ret==-1){
		perror ("sendmsg");
	}
}


static int aclfs_open (const char *path, int mode, int perm)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		//fprintf (stderr,"open :%s: %d %d\n",info.realpath,mode,perm);
		req.cmd = ACL_OPEN;
		req.mode = mode;
		req.perm = perm;
		aclfs_sendreq (req,len);
		ret = aclfs_receivefd();
	}
	return ret;
}

static int aclfs_mkdiranc (const char *path)
{
	int ret = __mkdir (path,0755);
	if (ret == -1){
		char tmp[strlen(path)+1];
		strcpy (tmp,path);
		char *pt = strrchr(tmp,'/');
		if (pt != NULL){
			*pt = '\0';
			if (aclfs_mkdiranc (tmp) != -1) ret = __mkdir (path,0755);
		}
	}
	return ret;
}


static int aclfs_chdir (const char *path)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		//fprintf (stderr,"chdir :%s: :%s: :%s:\n",path,info.virtpath,info.realpath);
		req.cmd = ACL_CHDIR;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
		if (ret != -1){
			// We have to set the new current dir in $HOME/n
			if (__chdir (info.virtpath) == -1){
				aclfs_mkdiranc (info.virtpath);
				__chdir (info.virtpath);
			}
		}
	}
	return ret;
}

static struct DIR * aclfs_opendir (const char *path)
{
	struct DIR *ret = NULL;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		req.cmd = ACL_OPENDIR;
		aclfs_sendreq (req,len);
		int fd = aclfs_receivefd();
		if (fd != -1){
			ret = (DIR*)fd;
		}
	}
	return ret;
}

static struct dirent *aclfs_readdir (struct DIR *dir)
{
	int fd = (int)dir;
	static struct dirent ent;
	struct dirent *ret = NULL;
	if (read (fd,&ent,sizeof(ent))==sizeof(ent)) ret = &ent;
	return ret;
}


static int aclfs_closedir (struct DIR *dir)
{
	int fd = (int)dir;
	return close (fd);
}
static int aclfs_rename (const char *old, const char *newp)
{
	int ret = DISPATCH_DONTCARE;
	#if 0
	ACLFS_INFO oldinfo;
	ACLFS_INFO newinfo;
	if (aclfs_translate (old,oldinfo)!=-1
		|| aclfs_translate (newp,newinfo)!=-1){
		ret = __libc_rename (oldinfo.realpath,newinfo.realpath);
	}
	#endif
	return ret;
}
static int aclfs_unlink (const char *path)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		// fprintf (stderr,"unlink :%s:\n",info.realpath);
		req.cmd = ACL_UNLINK;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}
static int aclfs_rmdir (const char *path)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		req.cmd = ACL_RMDIR;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}
static int aclfs_mkdir(const char *path, int mode)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (path,info,req,len)!=-1){
		req.cmd = ACL_MKDIR;
		req.mode = mode;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}

static int aclfs_stat (int ver, const char *fname, struct stat *buf)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		// fprintf (stderr,"stat :%s:\n",info.realpath);
		req.cmd = ACL_STAT;
		aclfs_sendreq (req,len);
		ret = aclfs_receive_stat(*buf);
	}
	return ret;
}
static int aclfs_lstat (int ver, const char *fname, struct stat *buf)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		// fprintf (stderr,"lstat :%s:\n",info.realpath);
		req.cmd = ACL_LSTAT;
		aclfs_sendreq (req,len);
		ret = aclfs_receive_stat(*buf);
	}
	return ret;
}

static int aclfs_chmod (const char *fname, mode_t mode)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		req.cmd = ACL_CHMOD;
		req.mode = mode;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}
static int aclfs_chown (const char *fname, uid_t uid, gid_t gid)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		req.cmd = ACL_CHOWN;
		req.mode = uid;
		req.perm = gid;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}
static int aclfs_lchown (const char *fname, uid_t uid, gid_t gid)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		req.cmd = ACL_LCHOWN;
		req.mode = uid;
		req.perm = gid;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}


static int aclfs_readlink (const char *fname, char *, size_t)
{
	/* #Specification: aclfs / readlink / not supported
		For now, we do not deal with symlinks in aclfs. We will
		revisit that once aclfs is proven (useful ?).
	*/
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		errno = EPERM;
		ret = -1;
	}
	return ret;
}
static int aclfs_utimes (const char *fname, struct timeval *tvp)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		req.cmd = ACL_UTIMES;
		if (tvp == NULL){
			req.mode = -1;
			req.perm = -1;
		}else{
			req.mode = tvp->tv_sec;
			req.perm = tvp->tv_usec;
		}
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}

static int aclfs_access (const char *fname, int type)
{
	int ret = DISPATCH_DONTCARE;
	ACLFS_INFO info;
	ACL_REQUEST req;
	int len;
	if (aclfs_translate (fname,info,req,len)!=-1){
		req.cmd = ACL_ACCESS;
		req.mode = type;
		aclfs_sendreq (req,len);
		ret = aclfs_receive();
	}
	return ret;
}

static DISPATCH m={
	NULL,
	aclfs_open,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	aclfs_stat,
	NULL,
	aclfs_lstat,
	aclfs_chdir,
	aclfs_opendir,
	aclfs_readdir,
	aclfs_closedir,
	aclfs_rename,
	aclfs_unlink,
	aclfs_rmdir,
	aclfs_mkdir,
	NULL,		// socket
	NULL,		// connect
	NULL,		// bind
	aclfs_chmod,
	aclfs_chown,
	aclfs_lchown,
	aclfs_readlink,
	aclfs_utimes,
	aclfs_access,
};

static REGISTER_DISPATCH register_m (m);


