#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <ctype.h>
#include <stdlib.h>
#include <dirent.h>
#include <signal.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <utime.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/socket.h>
#include <sys/un.h>
#include "aclfs.h"
#include "aclfsd.h"

#ifndef SCM_CREDENTIALS
	#define SCM_CREDENTIALS 2
	// Struct not available in headers yet, but supported by the kernel
	struct ACL_UCRED {
		int32_t	pid;
		int32_t	uid;
		int32_t	gid;
	};
#else
	typedef struct ucred ACL_UCRED;
#endif


static int aclfsd_server(const char *sockn)
{
	int ret = -1;
	unlink (sockn);
	int fd =  socket (AF_UNIX,SOCK_STREAM,0);
	if (fd == -1){
		perror("socket server");
	}else{
		struct sockaddr_un un;
		un.sun_family = AF_UNIX;
		strcpy (un.sun_path,sockn);
		if (bind(fd,(struct sockaddr*)&un,sizeof(un))==-1){
			perror("bind");
		}else{
			chmod (sockn,0666);
			int code = listen (fd,10);
			if (code == -1){
				perror ("listen");
			}else{
				ret = fd;
			}	
		}
	}
	return ret;
}

static int aclfsd_sendfd(int sock, int fd)
{
	char bufctl[sizeof(struct cmsghdr)+1*sizeof(int)];
	struct cmsghdr *c = (struct cmsghdr*)bufctl;
	c->cmsg_len = sizeof(bufctl);
	c->cmsg_level = SOL_SOCKET;
	c->cmsg_type = SCM_RIGHTS;
	int *fds = (int*)CMSG_DATA(c);
	fds[0] = fd;

	struct msghdr h;

	struct iovec iov[1];
	char buf[1000];
	strcpy (buf,"hellohello");
	iov[0].iov_base = buf;
	iov[0].iov_len = 10;

	h.msg_name = 0;
	h.msg_namelen = 0;
	h.msg_iov = iov;
	h.msg_iovlen = 1;
	h.msg_control = (void*)bufctl;
	h.msg_controllen = sizeof(bufctl);
	h.msg_flags = 0;
	int ret = sendmsg (sock,&h,0);
	if (ret==-1){
		perror ("sendmsg");
	}
	return ret;
}

static bool do_update = false;

/*
	Must reread its config
*/
static void aclfsd_sighup(int)
{
	do_update = true;
}


struct ACLFSD_CONF {
	char *volume;
	char *abspath;
};

static int nbdef=0;
static ACLFSD_CONF *conf;

static void aclfsd_readconf (const char *configf)
{
	if (conf != NULL){
		for (int i=0; i<nbdef; i++){
			free (conf[i].volume);
			free (conf[i].abspath);
		}
		free (conf);
		nbdef = 0;
	}
	FILE *fin = fopen (configf,"r");
	if (fin != NULL){
		char buf[1000];
		conf = (ACLFSD_CONF*)malloc (100*sizeof(ACLFSD_CONF));
		while (fgets(buf,sizeof(buf)-1,fin)!=NULL){
			if (buf[0] != '\0'){
				char *pt = buf;
				while (*pt > ' ') pt++;
				if (isspace(*pt)){
					*pt ++ = '\0';
					while (isspace(*pt)) pt++;
					char *end = pt + strlen(pt)-1;
					if (*end == '\n') *end = '\0';
					conf[nbdef].volume = strdup(buf);
					conf[nbdef].abspath = strdup(pt);
					nbdef++;
				}
			}
		}
		fclose (fin);
	}
}

static void aclfsd_sendfail (int fd)
{
	ACL_REPLY rep;
	rep.status = ACL_STATUS_FAIL;
	rep.errcode = EINVAL;
	write (fd,&rep,sizeof(rep));
}

/*
	Send an access violation error
*/
static void aclfsd_sendperm (int fd)
{
	ACL_REPLY rep;
	rep.status = ACL_STATUS_FAIL;
	rep.errcode = EPERM;
	write (fd,&rep,sizeof(rep));
}

static void aclfsd_sendok (int fd)
{
	ACL_REPLY rep;
	rep.status = ACL_STATUS_OK;
	rep.errcode = 0;
	write (fd,&rep,sizeof(rep));
}

static void aclfsd_sendstat (int fd, struct stat &st)
{
	struct {
		ACL_REPLY rep;
		struct stat st;
	}s;
	s.rep.status = ACL_STATUS_STAT;
	memcpy (&s.st,&st,sizeof(s.st));
	write (fd,&s,sizeof(s));
}

/*
	Convert the request to an absolute path by matching a volume
	definition. Send failure to the client if there is no matching
	volume or the request is invalid (.. in path)

	Return -1 if this is not a valid request.
*/
static int aclfsd_translate (
	int fd,
	ACL_REQUEST *req,
	char abspath[PATH_MAX],
	ACL_RIGHTS &rights,
	ACL_UCRED *uc)
{
	int ret = -1;
	char *path = req->paths;
	if (path[0] == '/') path++;
	char *pt = strchr(path,'/');
	if (pt != NULL)	*pt++ = '\0';
	for (int i=0; i<nbdef; i++){
		if (strcmp(conf[i].volume,path)==0){
			if (pt == NULL){
				strcpy (abspath,conf[i].abspath);
				aclrights_get (uc->uid,uc->gid,uc->pid,abspath,NULL,rights);
				ret = 0;
			}else if (strstr(pt,"..")==NULL){
				snprintf (abspath,PATH_MAX-1,"%s/%s",conf[i].abspath,pt);
				pt = strrchr(abspath,'/');
				*pt = '\0';
				aclrights_get (uc->uid,uc->gid,uc->pid,abspath,pt+1,rights);
				*pt = '/';
				ret = 0;
			}
			break;
		}
	}
	if (ret == -1){
		aclfsd_sendfail (fd);
	}
	return ret;
}


/*
	Read a request from one client
	Every request must contain a credential, to be added later
*/
static int aclfsd_doreq (int fd)
{
	char bufctl[sizeof(struct cmsghdr)+sizeof(ACL_UCRED)];
	memset (bufctl,-1,sizeof(bufctl));
	struct iovec iov[1];
	char buf[1000];
	iov[0].iov_base = buf;
	iov[0].iov_len = sizeof(buf);
	struct msghdr h;
	h.msg_name = 0;
	h.msg_namelen = 0;
	h.msg_iov = iov;
	h.msg_iovlen = 1;
	h.msg_control = bufctl;
	h.msg_controllen = sizeof(bufctl);
	struct cmsghdr *c = (struct cmsghdr*) bufctl;
	c->cmsg_len = CMSG_LEN(sizeof(ACL_UCRED));
	c->cmsg_level = SOL_SOCKET;
	c->cmsg_type = SCM_CREDENTIALS;

printf ("msg len %d %d\n",c->cmsg_len,h.msg_controllen);


	h.msg_flags = 0;
	int len = recvmsg (fd,&h,0);
	int ret = -1;
	ACL_UCRED *uc = (ACL_UCRED*)CMSG_DATA(c);
	printf ("len %d msg %d %d %04x uc %d %d %d\n",len
		,c->cmsg_level,c->cmsg_type,h.msg_flags
		,uc->pid,uc->uid,uc->gid);
	if (c->cmsg_level == SOL_SOCKET
		&& c->cmsg_type == SCM_CREDENTIALS){
		if (len>0){
			printf ("message recu flags %d len %d %d %u\n",h.msg_flags,len
				,h.msg_controllen,sizeof(bufctl));
			printf ("iov.len %d\n",iov[0].iov_len);
			printf ("msg_len %d msg_level %d msg_type %d\n",c->cmsg_len
				,c->cmsg_level
				,c->cmsg_type);
			ACL_REQUEST *req = (ACL_REQUEST*)buf;
			buf[len] = '\0';
			if (req->cmd == ACL_OPEN){
				printf ("open :%s:\n",req->paths);
				char abspath[PATH_MAX];
				ACL_RIGHTS rights;
				if (aclfsd_translate (fd,req,abspath,rights,uc)!=-1){
					printf ("open abspath :%s:\n",abspath);
					int handle = open (abspath,req->mode,req->perm);
					if (handle != -1){
						aclfsd_sendfd (fd,handle);
						close (handle);
					}else{
						aclfsd_sendfail(fd);
					}
				}
			}else if (req->cmd == ACL_STAT){
				printf ("stat :%s:\n",req->paths);
				char abspath[PATH_MAX];
				ACL_RIGHTS rights;
				if (aclfsd_translate (fd,req,abspath,rights,uc)!=-1){
					printf ("stat abspath :%s:\n",abspath);
					struct stat st;
					if (!rights.may_stat){
						aclfsd_sendperm(fd);
					}else if (stat(abspath,&st)==-1){
						aclfsd_sendfail(fd);
					}else{
						aclfsd_sendstat (fd,st);
					}
				}
			}else if (req->cmd == ACL_LSTAT){
				printf ("lstat :%s:\n",req->paths);
				char abspath[PATH_MAX];
				ACL_RIGHTS rights;
				if (aclfsd_translate (fd,req,abspath,rights,uc)!=-1){
					printf ("lstat abspath :%s:\n",abspath);
					struct stat st;
					if (!rights.may_stat){
						aclfsd_sendperm(fd);
					}else if (lstat(abspath,&st)==-1){
						aclfsd_sendfail(fd);
					}else{
						aclfsd_sendstat (fd,st);
					}
				}
			}else if (req->cmd == ACL_UNLINK){
				printf ("unlink :%s:\n",req->paths);
				char abspath[PATH_MAX];
				ACL_RIGHTS rights;
				if (aclfsd_translate (fd,req,abspath,rights,uc)!=-1){
					printf ("unlink abspath :%s:\n",abspath);
					if (unlink(abspath)!=-1){
						aclfsd_sendok (fd);
					}else{
						aclfsd_sendfail (fd);
					}
				}
			}else if (req->cmd == ACL_UTIMES){
				printf ("utimes :%s:\n",req->paths);
				char abspath[PATH_MAX];
				ACL_RIGHTS rights;
				if (aclfsd_translate (fd,req,abspath,rights,uc)!=-1){
					printf ("utimes abspath :%s: %d %d\n",abspath,req->mode,req->perm);
					int ok = -1;
					if (req->mode == -1 && req->perm == -1){
						ok = utimes (abspath,NULL);
					}else{
						struct timeval tvp;
						tvp.tv_sec = req->mode;
						tvp.tv_usec = req->perm;
						ok = utimes (abspath,&tvp);
					}
					if (ok!=-1){
						aclfsd_sendok (fd);
					}else{
						aclfsd_sendfail (fd);
					}
				}
			}else if (req->cmd == ACL_ACCESS){
				printf ("access :%s:\n",req->paths);
				char abspath[PATH_MAX];
				ACL_RIGHTS rights;
				if (aclfsd_translate (fd,req,abspath,rights,uc)!=-1){
					printf ("access abspath :%s: %d %d\n",abspath,req->mode,req->perm);
					int ok = -1;
					if (ok!=-1){
						aclfsd_sendok (fd);
					}else{
						aclfsd_sendfail (fd);
					}
				}
			}else if (req->cmd == ACL_RMDIR){
				printf ("rmdir :%s:\n",req->paths);
				char abspath[PATH_MAX];
				ACL_RIGHTS rights;
				if (aclfsd_translate (fd,req,abspath,rights,uc)!=-1){
					printf ("rmdir abspath :%s:\n",abspath);
					if (rmdir(abspath)!=-1){
						aclfsd_sendok (fd);
					}else{
						aclfsd_sendfail (fd);
					}
				}
			}else if (req->cmd == ACL_MKDIR){
				printf ("mkdir :%s:\n",req->paths);
				char abspath[PATH_MAX];
				ACL_RIGHTS rights;
				if (aclfsd_translate (fd,req,abspath,rights,uc)!=-1){
					printf ("mkdir abspath :%s:\n",abspath);
					if (mkdir(abspath,req->mode)!=-1){
						aclfsd_sendok (fd);
					}else{
						aclfsd_sendfail (fd);
					}
				}
			}else if (req->cmd == ACL_CHDIR){
				// chdir means, may chdir
				printf ("chdir :%s:\n",req->paths);
				char abspath[PATH_MAX];
				ACL_RIGHTS rights;
				if (aclfsd_translate (fd,req,abspath,rights,uc)!=-1){
					printf ("chdir abspath :%s:\n",abspath);
					struct stat st;
					if (stat(abspath,&st)!=-1 && S_ISDIR(st.st_mode)){
						aclfsd_sendok (fd);
					}else{
						aclfsd_sendfail (fd);
					}
				}
			}else if (req->cmd == ACL_OPENDIR){
				printf ("opendir :%s:\n",req->paths);
				char abspath[PATH_MAX];
				ACL_RIGHTS rights;
				if (aclfsd_translate (fd,req,abspath,rights,uc)!=-1){
					printf ("opendir abspath :%s:\n",abspath);
					DIR *dir = opendir (abspath);
					if (dir != NULL){
						int tbfd[2];
						if (pipe(tbfd)!=-1){
							aclfsd_sendfd (fd,tbfd[0]);
							if (fork()==0){
								struct dirent *ent;
								while ((ent=readdir(dir))!=NULL){
									write (tbfd[1],ent,sizeof(struct dirent));
								}
								closedir (dir);
								_exit (0);
							}
							close (tbfd[0]);
							close (tbfd[1]);
						}
						closedir(dir);
					}else{
						aclfsd_sendfail (fd);
					}
				}
			}else{
				aclfsd_sendfail(fd);
			}
			ret = 0;		
		}
	}
	return ret;
}
		



/*
	Process all request and connections. Infinite loop
*/
static void aclfsd_loop (int sock, const char *configf)
{
	int tbfd[1024];
	int nbfd = 0;
	fd_set fdin;
	FD_ZERO (&fdin);
	int maxfd = sock+1;
	FD_SET (sock,&fdin);
	signal (SIGHUP,aclfsd_sighup);
	signal (SIGCHLD,SIG_IGN);
	while (1){
		fd_set in;
		memcpy (&in,&fdin,sizeof(in));
		printf ("select maxfd %d\n",maxfd);
		if (select (maxfd,&in,NULL,NULL,NULL)!=-1){
			int orig_nbfd = nbfd;
			if (FD_ISSET (sock,&in)){
				struct sockaddr_un unc;
				size_t len = sizeof(unc);
				unc.sun_family = AF_UNIX;
				int s = accept (sock,(struct sockaddr*)&unc,&len);
				tbfd[nbfd++] = s;
				FD_SET (s,&fdin);
				if (s >= maxfd) maxfd = s + 1;
			}
			int *ptfd = tbfd;
			bool redo_fdin = false;
			for (int i=0; i< orig_nbfd; i++, ptfd++){
				int fd = *ptfd;
				if (FD_ISSET(fd,&in)){
					if (aclfsd_doreq (fd) == -1){
						close (fd);
						redo_fdin = true;
						*ptfd = -1;
					}
				}
			}
			if (redo_fdin){
				FD_ZERO (&fdin);
				maxfd = sock+1;
				FD_SET (sock,&fdin);
				ptfd = tbfd;
				int *copyfd = tbfd;
				for (int i=0; i<nbfd; i++, ptfd++){
					int fd = *ptfd;
					if (fd != -1){
						FD_SET (fd,&fdin);
						if (fd >= maxfd) maxfd = fd + 1;
						*copyfd++ = fd;
					}
				}
				nbfd = (int)(copyfd - tbfd);
			}
		}else if (do_update){
			do_update = false;
			aclfsd_readconf (configf);
		}
	}
}

int main (int argc, char *argv[])
{
	int ret = -1;
	if (argc == 1){
		{
			FILE *fout = fopen ("/var/run/aclfsd.pid","w");
			if (fout != NULL){
				fprintf (fout,"%d\n",getpid());
				fclose (fout);
			}
		}
		const char *configf = "/etc/aclfsd.conf";
		aclfsd_readconf(configf);
		int fd = aclfsd_server(ACLFS_SOCKET);
		if (fd != -1){
			aclfsd_loop(fd,configf);
		}
	}
	return ret;
}


